\ generic-scans.fs

VARIABLE (generic-scan-data)	\ will be in a field

' (generic-scan-data) 0
POINTER+OFFSET: generic-detail-xt    \ i.e. organ-A B-parameter score 
POINTER+OFFSET: generic-fetch-xt     \ i.e. noop @ df@
POINTER+OFFSET: generic-detail-type  \ i.e. type-int% type-df% type-int-addr%
POINTER+OFFSET: generic-fetch-type   \ i.e. type-int% type-df%
POINTER+OFFSET: generic-locality     \ i.e. nuc-local% spot-local%
POINTER+OFFSET: generic-loop-xt	     \ i.e. do-with-everybody do-everywhere

\ prescan data:
dup -rot

POINTER+OFFSET: prescan-count
POINTER+dfloatOFFSET: prescan-min
POINTER+dfloatOFFSET: prescan-max
POINTER+dfloatOFFSET: prescan-average	\ sum during prescan

rot over swap - CONSTANT prescan-data-size#

\ slice data
POINTER+OFFSET: slices
POINTER+OFFSET: out-of-range
POINTER+dfloatOFFSET: slice-min
POINTER+dfloatOFFSET: slices-range
maxaligned POINTER+OFFSET: slice-data-pointer	\ start address of slices
cell -
CONSTANT scan-data-size#			\ *without* slices
drop


\ Allocate and erase memory for scans, set the data pointer, initialize slices.
: allocate-scan-data ( pointer-addr slices -- )
    >r
    r@ cells  scan-data-size#  + allocate-clean
    swap !
    r> slices ! ;

(generic-scan-data) c-l allocate-scan-data


: fetch-int-as-float ( -- r )   generic-detail-xt @ EXECUTE s>f ;

: fetch-int-addr-as-float ( -- r )   generic-detail-xt @ EXECUTE @ s>f ;

: set-detail ( detail-xt detail-type -- )
    dup generic-detail-type !
    CASE
	type-int% OF
	    ['] fetch-int-as-float generic-fetch-xt !
	    type-int% generic-fetch-type !
	ENDOF
	type-int-addr% OF
	    ['] fetch-int-addr-as-float generic-fetch-xt !
	    type-int% generic-fetch-type !
	ENDOF

	ABORT
    ENDCASE
    generic-detail-xt ! ;

: set-locality ( locality-code -- )
    dup generic-locality !
    CASE
	nuc-local% OF
	    ['] do-with-everybody generic-loop-xt !
	    EXIT
	ENDOF
	spot-local%	OF
	    ['] do-everywhere generic-loop-xt !
	    EXIT
	ENDOF
	ABORT
    ENDCASE ;

: +count ( -- )   1 prescan-count +! ;

: prescan-process-float ( F: r -- )
    fdup  prescan-average  dup >r  df@ f+ r> df! \ sum up in prescan-average
    fdup  prescan-max  dup >r  df@  fmax  r> df! \ max
    prescan-min  dup >r  f@  fmin  r> f!	 \ min
    +count ;

: prescan-fetch&store-1 ( -- )
    generic-fetch-xt @ EXECUTE
    prescan-process-float ;

: init-prescan ( -- )
    0 prescan-count !
    0e0 prescan-average df!	\ used as sum during prescan
    +infinity prescan-min df!
    -infinity  prescan-max df! ;

: compute-average ( -- )
    prescan-average dup >r
    df@  fdup f0= IF  fdrop rdrop EXIT  THEN
    prescan-count @ s>f  f/  r> df! ;

: prescan ( -- )
    init-prescan
    ['] prescan-fetch&store-1  generic-loop-xt @  EXECUTE
    compute-average ;


: (set-int-var) ( xt -- )   type-int-addr%  set-detail ;

: set-nuc-int-var ( xt -- )   (set-int-var)  nuc-local% set-locality ;

: set-spot-int-var ( xt -- )   (set-int-var)  spot-local% set-locality ;


: (set-int-funct) ( xt -- )   type-int%  set-detail ;

: set-nuc-int-funct ( xt -- )   (set-int-funct)  nuc-local% set-locality ;

: set-spot-int-funct ( xt -- )   (set-int-funct)  spot-local% set-locality ;


: init-slicing ( -- )		\ data must be prescanned
    \ slices			\ set by allocate-scan-data

    
\    erase
    out-of-range off
    prescan-max df@
    prescan-min df@
    fdup slice-min df!		\ zooming???? DADA
    f-

    \ fix broken ranges
    generic-fetch-type @ CASE
	type-int% OF
	    \ range smaller than screen width; take a one to one relationship
	    c-l 1- s>f fmax
	ENDOF
	type-df% OF
	    fdup f0= IF  fdrop  1e0  THEN
	ENDOF
	ABORT
    ENDCASE

    slices-range df! ;

: n'th-slice ( n -- n'th-count-addr )   cells  slice-data-pointer  + ;

: data-to-slice ( F: r -- )
    slice-min df@ f-		( F: offset-from-min )
    slices-range df@ f/
    slices @ s>f f* f>s		( slice )

    dup 0< IF			\ below range
	1 out-of-range +!
	drop 0
    THEN

    dup slices @ 1- > IF	\ above range
	1 out-of-range +!
	drop
	slices @ 1-
    THEN

    1 swap n'th-slice +! ;

: fetch&slice-1 ( -- )   generic-fetch-xt @ EXECUTE  data-to-slice ;

: slice-up ( -- )
    init-slicing
    ['] fetch&slice-1  generic-loop-xt @  EXECUTE ;

: do-scan ( -- )   prescan slice-up ;

true [if] \ testing

    : .xt ( xt -- )
	dup 0= IF ." uninitialized xt" drop EXIT THEN
	xt>string type ;

    : .scandata ( -- )
	generic-detail-xt @	cr ." generic-detail-xt    " .xt
	generic-fetch-xt @	cr ." generic-fetch-xt     " .xt
	generic-detail-type @   cr ." generic-detail-type  "
	var-type-string type
	generic-fetch-type @	cr ." generic-fetch-type   "
	var-type-string type
	generic-locality @	cr ." generic-locality     "
	locality-string type
	generic-loop-xt @	cr ." generic-loop-xt      " .xt
	prescan-count @		cr ." prescan-count        " .
	prescan-min df@		cr ." prescan-min          " f.
	prescan-max df@		cr ." prescan-max          " f.
	prescan-average df@	cr ." prescan-average      " f.
	slice-data-pointer      cr ." slice-data-pointer   " .
	slices @		cr ." slices               " .
	out-of-range @		cr ." out-of-range         " .
	slice-min df@		cr ." slice-min            " f.
	slices-range df@        cr ." slices-range         " f.

	cr
	slices @ 0 DO   i n'th-slice @ .   LOOP
    ;

    : ttt
	page
	['] organ-A set-nuc-int-var
	prescan
	do-scan
	.scandata
	;

[then]
