\ assertions.fs
\ 	$Id: assertions.fs,v 1.2 2009-10-15 16:06:47 f Exp $

\ Words related to assert state of a recorded session or a benchmark:

\ This words are mainly for the creation of benchmarks.
\ Call assert state from record/playback menu for result validation of a
\ recorded session.  Or if you want to add it later, put 'assert-state-entry'
\ into the record file at the appropriate place and replace that by the
\ produced FORTH code to check state when playing it back.


\ Words that will be interpreted when playing back assert statements
\ created by 'assert-variable-entry' and similar.
\ As it get's the xt of the variable it can give error diagnostics.
: assert@= ( n xt -- flag )
    swap asserted-bits swap
    2dup EXECUTE @ asserted-bits = IF
	2drop TRUE
	EXIT
    THEN

    dup xt>string cr type
    ."  is " EXECUTE @ asserted-bits . ." instead of " .
    FALSE ;

: assert2@= ( d xt -- flag )
    >r
    d-asserted-bits 2dup
    r@ EXECUTE 2@ d-asserted-bits d= IF
	2drop rdrop TRUE
	EXIT
    THEN

    r@ xt>string cr type
    ."  is " r> EXECUTE 2@ d-asserted-bits swap . . ." instead of " swap . .
    FALSE ;

: assert@u= ( u xt -- flag )
    swap asserted-bits swap
    2dup EXECUTE @ asserted-bits = IF
	2drop TRUE
	EXIT
    THEN

    dup xt>string cr type
    ."  is " EXECUTE @ asserted-bits u. ." instead of " u.
    FALSE ;

: assert2@u= ( d xt -- flag )
    >r
    d-asserted-bits 2dup
    r@ EXECUTE 2@ d-asserted-bits d= IF
	2drop rdrop TRUE
	EXIT
    THEN

    r@ xt>string cr type
    ."  is " r> EXECUTE 2@ d-asserted-bits swap u. u.
    ." instead of " swap u. u.
    FALSE ;

: assert-variable-entry ( xt -- )
    dup EXECUTE @ asserted-bits
    num>string			cat2out
    s"  	' "		cat2out
    xt>string			cat2out
    s"  	assert@=  AND"	cat-and-out ;

: assert-2variable-entry ( xt -- )
    dup EXECUTE 2@ d-asserted-bits
    swap num>string		cat2out
    bl				char2out
    num>string			cat2out
    s"  	' "		cat2out
    xt>string			cat2out
    s"  	assert2@=  AND"	cat-and-out ;

: assert-u-variable-entry ( xt -- )
    dup EXECUTE @ asserted-bits
    u-num>string		cat2out
    s"  	' "		cat2out
    xt>string			cat2out
    s"  	assert@u=  AND"	cat-and-out ;

: assert-u-2variable-entry ( xt -- )
    dup EXECUTE 2@ d-asserted-bits
    swap u-num>string		cat2out
    bl				char2out
    u-num>string		cat2out
    s"  	' "		cat2out
    xt>string			cat2out
    s"  	assert2@u=  AND"  cat-and-out ;

\ Word that will be interpreted when playing back assert statements
\ created by 'assert-function-entry'.  
\ As it get's the xt of the function it can give error diagnostics.
: assert-do= ( n do-xt -- flag )
    swap asserted-bits swap
    2dup EXECUTE asserted-bits = IF
	2drop TRUE
	EXIT
    THEN

    dup xt>string cr type
    ."  gives " EXECUTE asserted-bits . ." instead of " .
    FALSE ;

\ Make assert entry for a word giving one output parameter:
: assert-function-entry ( xt -- )
    dup EXECUTE asserted-bits
    num>string			cat2out
    s" 	' "			cat2out
    xt>string			cat2out
    s" 	assert-do=  AND"	cat-and-out ;

: assert-do-u= ( u do-xt -- flag )
    swap asserted-bits swap
    2dup EXECUTE asserted-bits = IF
	2drop TRUE
	EXIT
    THEN

    dup xt>string cr type
    ."  gives " EXECUTE asserted-bits u. ." instead of " u.
    FALSE ;

\ Make assert entry for a word giving one unsigned output parameter:
: assert-u-function-entry ( xt -- )
    dup EXECUTE asserted-bits
    u-num>string		cat2out
    s" 	' "			cat2out
    xt>string			cat2out
    s" 	assert-do-u=  AND"	cat-and-out ;


\ Produce FORTH code to check the current state of brew.
\ When the code is run it assures the same state as when it was produced.
\ Used to check validity of benchmark results.
: assert-state-entry ( -- )
    page cr

    (outfile-id) @			( old-outfile-id )

    cr ." Saving code to assert current brew state to: "
    NOT-recording? IF
	file-names-length# stringbuf-open >r
	tmp-dir			r@ cat
	s" ASSERTIONS.fs"	r@ cat
	r@ string@ type cr
	r@ string@ w/o CREATE-FILE+  r> stringbuf-close
	IF
	    bell
	    cr ." assert-state-entry: Couldn't create file."
	    2000 ms
	    drop
	    EXIT
	THEN
	set-outfile
	true			\ flag: do close file when done
    ELSE
	(record-file-name) string@ type cr
	(record-file-id) @
	dup set-outfile
	record-brew-changes
	false			\ flag: do not close file
    THEN

    cr
    cr ." You can give a comment:" cr
    c-l stringbuf-open
    dup accept>stringbuf cr

    out-line
    s" \ assert-state-entry: Assert expected results:" cat-and-out
    s" true"	cat-and-out
    which-random-seed CASE		( old-id xt )
	1 OF assert-u-variable-entry ENDOF
	2 OF assert-u-2variable-entry ENDOF
	true ABORT" assert-state-entry: Error saving random seed."
    ENDCASE

    ['] step		assert-variable-entry
    ['] cloned		assert-variable-entry
    ['] living		assert-variable-entry
    ['] nuc-do-cost	assert-variable-entry
    ['] code-price	assert-variable-entry
    ['] (mutated-max)	assert-variable-entry
    ['] compiled-genes	assert-variable-entry

    ['] world-checksum	assert-u-function-entry

    ['] nucs-checksum	assert-u-function-entry

    s" cr"				cat-and-out
    dup string@ dup IF
	s" .( "				cat2out
	( addr count )			cat2out
	s" )"				cat-and-out
    ELSE 2drop THEN
    stringbuf-close
	
    s" [IF]"				cat-and-out
    s"     .( 	Result is valid. )"	cat-and-out
    s" [ELSE]"				cat-and-out
    s"     bell .( 	Unexpected result. Not comparable to other systems! )"
    cat-and-out
    s"     playing-bench? 0= [IF]"	cat-and-out
    s"         wait"			cat-and-out
    s"     [THEN]"			cat-and-out
    s" [THEN]"				cat-and-out
    out-line

    ( flag ) IF close-outfile THEN

    set-outfile
    1500 wait-until ;
