\ brew-basics.fs
\ 	$Id: brew-basics.fs,v 1.117 2009-10-15 16:58:02 f Exp $	

\ doing some very basic definitions first makes life easier...

\ ****************************************************************
\ file dependencies:
\ *Inside* brew this dependencies are met anyway...

\ Make sure basics.fs (which defines the mark BASICS.FS) is loaded:
bl parse BASICS.FS dup pad c! pad char+ swap chars move pad find nip 0=
[IF]	s" basics.fs" INCLUDED		[THEN]

s" lists.fs"		REQUIRED
s" listed-masks.fs"	REQUIRED
s" compile-options.fs"	REQUIRED
s" simple-stringbuf.fs"	REQUIRED
s" advanced.fs"		REQUIRED
s" display.fs"		REQUIRED
s" program-basics.fs"	REQUIRED
s" messages.fs"		REQUIRED


\ ****************************************************************
\ Compile options:
\ 	preset-run-mode
\       color-offset

\ ****************************************************************

decimal


VARIABLE run-mode	run-mode off		\ mode like recording and such
[DEFINED] preset-run-mode [IF]
    preset-run-mode run-mode !
[THEN]

LIST: run-mode-masks
run-mode-masks
0
LISTED-MASK: mutating
LISTED-MASK: no-code-cost	\ for 'gene-follow-mute'
LISTED-MASK: writing-code	\ writing code *only*
LISTED-MASK: recording
LISTED-MASK: write-diff	\ writing diff files (for recording or logging)
LISTED-MASK: playback
LISTED-MASK: linear-mode	\ should become world local
LISTED-MASK: elitism		\ should become world local
LISTED-MASK: making-bench
LISTED-MASK: playing-bench
LISTED-MASK: store-genomes	\ put compiled genomes and usage in a list
LISTED-MASK: compiled		\ mutated genome is already compiled (rebirth)
LISTED-MASK: redisplaying
LISTED-MASK: defining-rectangle	\ defining a spot rectangle
LISTED-MASK: defining-bar-range	\ defining a column bar range
2drop

' run-mode run-mode-masks compile-listed-?-and-!

: NOT-recording? ( -- flag )  run-mode @ recording and 0= ;

: world-mode? ( -- flag )   run-mode @ linear-mode and  0= ;

: elitism-off ( -- )  run-mode dup @  [ elitism invert ] literal and swap ! ;

\ do something not recording it
: not-recording ( xt -- )
    run-mode dup @
    dup >r			\ preserve run-mode
    recording invert and swap !
    EXECUTE
    r> run-mode ! ;

64 CONSTANT max-stack-effect	\ no genom should ever go beyond that
max-stack-effect cell mod [IF]
    cr cr .( Compile time error: 'max-stack-effect CELL MOD' must be zero! ) cr
    QUIT
[THEN]

DEFER <food> ( -- addr )		\ gives addr where the food is

VARIABLE (manually-selected-cell)	(manually-selected-cell) off


\ Display:

VARIABLE display-switch			\ switches at run time

\ If you define (uncomment) 'never-use-colors' in 'compile-options.fs'
\ there will be no colors used. (Most color stuff will still get compiled
\ in though for the sake of simplicity.)

\ bit masks for 'display-switch':
LIST: display-switch-masks
display-switch-masks
0
LISTED-MASK: spot-display-on		\ spot display on
LISTED-MASK: spot-foreground-coloring	\ spot foreground color on
LISTED-MASK: spot-background-coloring	\ spot background color on
LISTED-MASK: step-display-on		\ step statistic info, nuc/world scans
LISTED-MASK: step-foreground-coloring	\ spot foreground color on
LISTED-MASK: step-background-coloring	\ spot background color on
\ snapshot bits only have an influence if both ordinary display types are off:
LISTED-MASK: step-snapshots		\ switches step snapshots on
LISTED-MASK: spot-snapshots		\ switches spot snapshots on
LISTED-MASK: continuous-display-used	\ does *not* switch continuous display
LISTED-MASK: scan-display-used		\ does *not* switch scan display
LISTED-MASK: text-display-used		\ does *not* switch text display
2drop

' display-switch display-switch-masks compile-listed-?-and-!

spot-display-on step-snapshots or display-switch !

\ these switches work on *spot* display: (currently)
: toggle-foreground-colorizing ( -- )
    display-switch dup @ spot-foreground-coloring xor swap ! ;
: toggle-background-colorizing ( -- )
    display-switch dup @ spot-background-coloring xor swap ! ;

: toggle-colorizing ( -- )
    display-switch dup @		( switch-addr switch-bitmask )

    dup spot-display-on and IF		\ spot display on
	[ spot-foreground-coloring spot-background-coloring + ] literal >r
	dup r@ and r@ = IF			\ full colored spot display?
	    spot-foreground-coloring xor	\ colors off
	    spot-background-coloring xor
	ELSE r@ or THEN rdrop			\ colors on
    THEN

    dup step-display-on and IF		\ step display on
    THEN

    swap ! ;

\ Transform any given number in an allowed color code.
[UNDEFINED] 2-color [IF] \ could be defined system dependent.

\ Assuming continuous range from zero (or 'color-offset') upwards.
\ compile option: 'color-offset' not used, never tested.
\ could give problems in other places (data visualization).
: 2-color ( col -- col' )
    colors mod
    dup 0< IF colors + THEN

    [DEFINED] color-offset [IF]	\ color-offset is the lowest color code
	[ color-offset ] literal +  \ for systems where colors don't start at 0
    [THEN]
;
[THEN]

VARIABLE background-color-xt	' black background-color-xt !
: <background-color> ( -- col )		background-color-xt @ EXECUTE ;

VARIABLE foreground-color-xt	' white foreground-color-xt !
: <foreground-color> ( -- col )		foreground-color-xt @ EXECUTE ;

: set-colors ( -- )	\ for spot display
    spot-display-on? IF
	spot-foreground-coloring? IF
	    <foreground-color> color-foreground
	THEN
	spot-background-coloring? IF
	    <background-color> color-background
	THEN
    THEN ;

LIST: x>bg-color		\ words that set color depending on something
LIST: x>fg-color		\ words that set color depending on something

\ first the trivial cases:
' black		dup x>bg-color >list	x>fg-color >list
' red		dup x>bg-color >list	x>fg-color >list
' green		dup x>bg-color >list	x>fg-color >list
' brown		dup x>bg-color >list	x>fg-color >list
' blue		dup x>bg-color >list	x>fg-color >list
' magenta	dup x>bg-color >list	x>fg-color >list
' cyan		dup x>bg-color >list	x>fg-color >list
' white		dup x>bg-color >list	x>fg-color >list
' default-color	dup x>bg-color >list	x>fg-color >list

[DEFINED] default-foreground [IF]
    ' default-foreground x>fg-color >list
[THEN]
[DEFINED] default-background [IF]
    ' default-background x>bg-color >list
[THEN]

: paint-background ( -- )
    spot-background-coloring? IF
	<background-color> color-background
    THEN ;

: set-default-colors ( -- )	\ same, but checks if needed
    display-switch @
    dup spot-foreground-coloring and IF default-foreground THEN
	spot-background-coloring and IF default-background THEN ;

\ brews file tree:
: genes-dir ( -- addr count )	      s" genes/" inputs-sub-dir ;
: experiments-dir ( -- addr count )   s" experiments/" inputs-sub-dir ;
: individuals-dir ( -- addr count )   s" individuals/" inputs-sub-dir ;

: rec-play-dir ( -- addr count )	s" rec-play/"	out-sub-dir ;
: log-dir ( -- addr count )		s" log/"	out-sub-dir ;
: FORTH-dir ( -- addr count )		s" FORTH/"	out-sub-dir ;


\ Defining menu-id's:
1
ENUM: main-sceen-id
ENUM: nuc-menu-id
drop


\ Defining scan locality type's:
LIST: locality-types
locality-types
0
LISTED-ENUM: unknown-locality%
LISTED-ENUM: global-locality%
LISTED-ENUM: world-local%
LISTED-ENUM: spot-local%
LISTED-ENUM: nuc-local%
\ LISTED-ENUM: nuc-and-spot-local% \ inhabited spots and nucs
\ LISTED-ENUM: world-local%
2drop

: locality-string ( locality-code -- addr count )
    locality-types listed-enum>string ;


\ Defining THROW codes:
1
[UNDEFINED] |menu-input-error [IF]
    ENUM: |menu-input-error
[ELSE] 1+ [THEN]
ENUM: |genome-too-long
ENUM: |playback-quit
ENUM: |stack-symbols-mismatch
drop


\ Defining item masks for nuc and spot vars:
LIST: item-masks

\ Define bitmasks named |A |B |C ...
: define-item-masks ( u -- )
    >r
    item-masks 0
    r> 0 ?DO
	s" LISTED-MASK: |X"		\ the 'X' will be overwritten...
	2dup 1- +			\ address of the 'X'
	i [char] A +  swap c!
	EVALUATE
    LOOP
    2drop ;
decimal 32 define-item-masks

\ A word to use the old integer scales on floats:
: f*/ ( F: r  D: n1 n2 -- F: r*n1/n2 )   s>f f/  s>f f* ;


\ Define variable types:
LIST: variable-types

variable-types 0
LISTED-ENUM: type-unknown%
LISTED-ENUM: type-int%
\ LISTED-ENUM: type-double%
LISTED-ENUM: type-df%
LISTED-ENUM: type-int-addr%	\ pointer to int
\ LISTED-ENUM: type-double-addr%	\ pointer to double int cell
LISTED-ENUM: type-df-addr%	\ pointer to dfloat
\ LISTED-ENUM: type-xt%		\ xt (general)
\ LISTED-ENUM: type-xt(-n)%	\ xt ( -- n )
\ LISTED-ENUM: type-xt(-a)%	\ xt ( -- a )
\ LISTED-ENUM: type-xt(-f)%	\ xt ( -- flag )
\ LISTED-ENUM: type-bitmap%
2drop

: var-type-string ( var-type-code - addr count )
    variable-types listed-enum>string ;

: is-int? ( type-code -- flag )
    CASE
	type-int-addr%	OF  TRUE  EXIT  ENDOF
	type-int%	OF  TRUE  EXIT  ENDOF
    ENDCASE
    FALSE ;

: is-dfloat? ( type-code -- flag )
    CASE
	type-df-addr%	OF  TRUE  EXIT  ENDOF
	type-df%	OF  TRUE  EXIT  ENDOF
    ENDCASE
    FALSE ;

\ ****************************************************************
\ Comparing checksums between 32 and 64 bits

\ Running old 32bit benchmarks on 64bit system can let assertions
\ fail even if the result is correct. So we switch to a mode, where
\ only 32 bits of the asserted values are checked. For some values
\ i.e. random seeds one must use *unsigned* values to make them look
\ the same on 32 and 64 bit systems. No need to change the old negative
\ 32bit values in old recorded benchmarks, btw.

VARIABLE (assert-bits)		\ bitmask for the bits to assert

: asserted-bits ( n -- n' )   (assert-bits) @ and ;

: d-asserted-bits ( d -- d' )	\ not quite sure how to treat this one...
    swap asserted-bits  swap asserted-bits ;

: assert-all-bits! ( -- )   (assert-bits) on ;
\ assert-all-bits!		\ defaults to the machine in use

hex
: assert-32-bits! ( -- )   ffffffff (assert-bits) ! ;
assert-32-bits!			\ conservative default right for now...

cell 7 > [IF]
    \ I leave this undefined on 32bit systems, so it will fail when a
    \ benchmark explicitly asks for 64bit assertions on a 32 bit system.
    : assert-64-bits! ( -- )   ffffffffffffffff (assert-bits) ! ;
[THEN]
decimal
