\ early-compile-hook.fs
\ $Id: early-compile-hook.fs,v 1.7 2009-10-20 18:09:25 f Exp $
\ ****************************************************************

\ This hook is for adaption to special cases, like running Anton Ertls
\ appbench suite. In normal brew compilation it should be a noop.

\ This code is run very early in the compilation process before the
\ system is properly set up. So don't rely on words like [DEFINED] or
\ [REQUIRED] and friends to be defined here.

\ ****************************************************************
\ CREATE no-term, no-key and random-only-32bit on a couple of benchmarks:
\
\ Anton Ertls appbench suite needs 'noterm' to bypass some terminal words.
\ I set it here for a couple of benchmarks including 'transit-11-bench-A'
\ used by appbench. Renamed to no-term to allow check for an existing word.
\
\ As appbench suppresses brews output it does not report an interrupt of
\ the bench by pressing a key as a failure. So I disregard key input while
\ running one of these benchmarks.
\
\ The old benchmarks used 32bit random generators. So 'transit-11-bench-A'
\ only succeeds on 64bit with an adaption of the random generators to 32bit.
\ 
\ Sanity checks:
\ Let's hope noone of the used FORTH system has such a word defined!
found? no-term [IF]
    cr
    cr .( WARNING: 'no-term' was already defined when starting brew. )
    cr .(          Brews user interface  will *not* work like this. )
    cr .(          See file 'early-compile-hook.fs'. )
    cr .(          Please do report that to the author. )
    cr
    2000 ms
[THEN]
found? no-key [IF]
    cr
    cr .( WARNING: 'no-key' was already defined when starting brew. )
    cr .(          Brew will *not* react on key input. )
    cr .(          See file 'early-compile-hook.fs'. )
    cr .(          Please do report that to the author. )
    cr
    5000 ms
[THEN]

found? startup-bench
found? transit-11-bench-A or
found? transit-12-bench or
[IF]
    CREATE no-term	\ bypass some terminal words
    CREATE no-key	\ disregard key input
    : key? ( -- false )    false ;
    : ekey? ( -- false )   false ;

    \ Compile switch  random-only-32bit
    \ The old benchmarks rely on 32bit only random generators.
    \ At least 'transit-11-bench-A' gives valid results on 64bit systems
    \ with this compile time switch.
    \ In some other benchmarks it is not enough to set this switch,
    \ as there are other issues elsewhere.
    1 cells 4 > [IF]
	CREATE random-only-32bit
	CREATE 32bit-64	   \ adaptions needed to run 32bit benchmarks on 64bit
    [THEN]

    \ Creating the file tree could possibly give problems,
    \ so I switch it off
    CREATE no-create-file-tree
[THEN]

\ ****************************************************************

found? brew-crash-test  1 cells 4 >  and [IF]
    CREATE random-only-32bit
    CREATE 32bit-64	   \ adaptions needed to run 32bit benchmarks on 64bit
[THEN]

[DEFINED] 32bit-64 [IF] \ Compile words to deal with 32bit data on 64bit system
    s" 32bit64bit.fs" INCLUDED
[THEN]
