\ file-handling.fs
\ 	$Id: file-handling.fs,v 1.5 2009-10-11 12:10:23 f Exp $

s" basics.fs"		REQUIRED
s" stringbuf-0.4.fs"	REQUIRED
s" advanced.fs"		REQUIRED
s" display.fs"		REQUIRED
s" file-tree.fs"	REQUIRED
s" identity.fs"		REQUIRED



\ Change a file name in a buffer.
\ If the user changes it close the old file and clear the id variable.
\ Does *not* open or create the new one.
: change-handled-file ( addr-of-id-variable name-handle -- )
    dup string@				\ old name
    dup stringbuf-open >r  r@ cat	\ kept in a buffer

    dup accept>stringbuf	\ let the user change the name
    dup string@  r@ string@  compare IF \ changed?
	over @ dup IF			\ old file was opened?
	    CLOSE-FILE			\ close it
	    IF
		bell
		cr ." change-file: Couldn't close file "
		r@ string@ type cr
		key drop		\ could be something serious
	    THEN
	    over off			\ and set id to zero
	ELSE drop THEN
    THEN

    r> stringbuf-close
    2drop ;


\ Keep track of temporary files for removal on exit:
VARIABLE (tmp-file-list-id)
VARIABLE tmp-files		tmp-files off	\ flag (counter)

\ tmp-files-list  will be created on first usage to avoid it's creation
\ on benchmarks.
TRUE [IF] \ Individual tmp file list files for each brew incarnation.
    : tmp-files-list ( -- handle )	\ please do close buffer.
	tmp-dir s" tmp-files" string-cat
	identity-string third cat ;

    : ?create-tmp-file-list ( -- )
	tmp-files @ IF  EXIT  THEN

	tmp-files-list dup string@ w/o CREATE-FILE IF
	    drop bell
	    cr ." Could not create file " over string@ type clear-line-to-end
	    2000 ms
	ELSE
	    (tmp-file-list-id) !
	THEN
	stringbuf-close ;
	
[ELSE] \ only one tmp file list: one brew could remove the others tmp files...
    : tmp-files-list ( -- handle )	\ please do close buffer.
	tmp-dir s" tmp-files" string-cat ;

    : ?create-tmp-file-list ( -- )
	tmp-files @ IF  EXIT  THEN

	tmp-files-list dup string@ r/w OPEN-FILE  IF	\ try to use old file
	    drop
	    dup string@ r/w CREATE-FILE IF		\ re-create file
		bell drop
		cr ." Could not create file " over string@ type
		clear-line-to-end
		2000 ms
	    ELSE
		(tmp-file-list-id) !
	    THEN
	ELSE	\ no file exists
	    dup file-size drop third reposition-file drop
	    (tmp-file-list-id) !
	THEN
	stringbuf-close ;
[THEN] \ tmp file list separated or not for each brew

\ Put a file name into file  (tmp-file-list-id) :
: to-tmp-file-list ( addr count -- )
    ?create-tmp-file-list
    (tmp-file-list-id) @ write-line drop
    (tmp-file-list-id) @ flush-file drop
    1 tmp-files +! ;

\ Create a file, on success put the name into  (tmp-file-list-id)  for later
\ removal, and in  (last-tmp-file-name)  for reference:
\ The id is in  (last-tmp-file-id)
file-names-length# STRINGBUF-HANDLE: (last-tmp-file-name)
VARIABLE (last-tmp-file-id)	(last-tmp-file-id) off
: CREATE-tmp-FILE ( c-addr u wfam - wfileid wior )
    third third 2>r
    CREATE-FILE dup 0= IF
	2r@ to-tmp-file-list
	2r@ (last-tmp-file-name) string!
	over (last-tmp-file-id) !
    THEN
    2rdrop ;

: reopen-last-tmp-file ( -- wior )
    (last-tmp-file-name) string@ r/w OPEN-FILE
    dup IF
	(last-tmp-file-id) off		\ programmers aesthetics...
    ELSE
	swap (last-tmp-file-id) !	\ on success set  (last-tmp-file-id)
    THEN ;

: ?remove-tmp-files ( -- )
    tmp-files @ 0= IF  EXIT  THEN

    (tmp-file-list-id) @ CLOSE-FILE drop
    tmp-files-list
    dup string@ r/o OPEN-FILE IF
	bell drop
	cr ." Could not open file " dup string@ type clear-line-to-end
	stringbuf-close
	2000 ms EXIT
    THEN
    cr ." Removing tmp files." clear-line-to-end
    file-line-max# allocate ABORT" remove-tmp-files: Could not allocate."
    >r
    BEGIN
	r@ file-line-max# third READ-LINE
	ABORT" remove-tmp-files: Could not read-line."
    WHILE
	r@ swap DELETE-FILE drop
    REPEAT
    drop
    CLOSE-FILE drop
    r> free ABORT" remove-tmp-files: Could not free."
    dup string@ DELETE-FILE drop	\ remove file list itself
    stringbuf-close ;

\ Keep trace of produced output files:
: created-files-list ( -- handle )	\ please do close buffer.
    tmp-dir s" created-files" string-cat
    identity-string third cat ;

VARIABLE created-files		created-files off	\ counter
VARIABLE (created-files-list-ID)	(created-files-list-ID) off

\ Avoid  created-files-list  in benchmarks, don't create file until first use:
: ?create-created-files-list ( -- )	\ Created on first usage
    created-files @ IF  EXIT  THEN
    
    created-files-list dup string@ w/o CREATE-FILE IF
	bell
	cr ." Could not create file " over string@ type clear-line-to-end
	2000 ms
    ELSE
	(created-files-list-ID) !
	dup string@ (created-files-list-ID) @ WRITE-LINE drop
	(created-files-list-ID) @ FLUSH-FILE drop
    THEN stringbuf-close ;

: to-created-file-list ( addr count -- )
    ?create-created-files-list
    (created-files-list-ID) @ write-line drop
    (created-files-list-ID) @ flush-file drop
    1 created-files +! ;

\ Create a file and put it in  created-files-list , count in  created-files :
: CREATE-FILE+ ( addr count wfam - wfileid wior )
    third third 2>r
    CREATE-FILE
    dup 0= IF  2r@ to-created-file-list  THEN
    2rdrop ;

\ Open a file and put it in  created-files-list , count in  created-files :
: OPEN-FILE+ ( addr count wfam - wfileid wior )
    third third 2>r
    OPEN-FILE
    dup 0= IF  2r@ to-created-file-list  THEN
    2rdrop ;

\ Build name of a backup file and return buffer:
: backup-file-name ( addr count -- handle )	\ Please close buffer.
    string!!
    s" .bak" third cat ;

\ Create a backup file and put it in  created-files-list ,
\ count in  created-files :
: create-backup-file+ ( addr count -- )
    2dup backup-file-name >r
    r@ string@
    2dup to-tmp-file-list
    clone-file
    r> stringbuf-close ;

\ Compare two files and return a flag which is TRUE if there's a difference:
: compare-files ( addr1 count1 addr2 count2 -- diff-flag )
    2dup r/o OPEN-FILE IF
	bell
	cr ." compare-files: Could not open-file  " type
	true ABORT" compare-files: ABORTED."
    ELSE
	>r 2drop	( addr1 count1  r: file-id-2 )
    THEN
    2dup r/o OPEN-FILE IF
	bell
	cr ." compare-files: Could not open-file  " type
	true ABORT" compare-files: ABORTED."
    ELSE
	nip nip	( file-id-1  r: file-id-2 )
    THEN

    dup FILE-SIZE
    ABORT" compare-files: Could not determine file-size on first file."
    r@ FILE-SIZE
    ABORT" compare-files: Could not determine file-size on second file."
    2dup d>s >r
    d= IF			\ same size
	( file-id-1  r: file-id-2 file-size )
	r@ stringbuf-open
	dup buffer-data-addr	( f-id-1 handle buffer-addr-1  r: f-id-2 size )
	r@ fourth read-file
	ABORT"  compare-files: Could not read-file #1."
	over buffer-data-addr swap	( id1 h1 addr1 count1  r: f-id-2 size )

	r@ stringbuf-open dup buffer-data-addr
	r> r@ read-file
	ABORT"  compare-files: Could not read-file #2."
	over buffer-data-addr swap
	rot >r		( id1 h1 addr1 count1 a2 c2  r: f-id-2 h2 )
	compare >r
	stringbuf-close r> r> stringbuf-close	( f-id1 flag  r: f-id-2 )
	0<>
    ELSE  rdrop TRUE  THEN	\ size differs
    swap CLOSE-FILE
    ABORT" compare-files: Could not close-file #1."
    r> CLOSE-FILE
    ABORT" compare-files: Could not close-file #2." ;

\ Compare a file to it's backup file and return difference flag:
: compare-to-backup ( addr count -- flag )
    2dup backup-file-name >r
    r@ string@ compare-files
    r> stringbuf-close ;

\ Create a named file, abort on error.
\ Put the name in  created-files-list  and count in  created-files :
: create-named-file ( addr count wfam -- fid )
    third third 2>r
    CREATE-FILE+
    dup IF
	bell cr ." create-named-file: Error creating "
	2r@ type cr
	THROW
    THEN drop
    2rdrop ;

\ Open a named file, abort on error.
: open-named-file ( addr count wfam -- fid )
    third third 2>r
    OPEN-FILE
    dup IF
	bell cr ." open-named-file: Error opening "
	2r@ type cr
	THROW
    THEN drop
    2rdrop ;

: ?list-created-files ( -- )
    created-files @ 0= IF
	cr ." No output files produced.    "  clear-line-to-end
	cr clear-line-to-end	0 at-x
	EXIT
    THEN

    (created-files-list-ID) @ CLOSE-FILE drop
    created-files-list dup string@ r/w OPEN-FILE IF
	drop
	cr  ." File  " dup string@ type ."  not found." clear-line-to-end
	stringbuf-close
	EXIT
    THEN

    cr
    cr ." You created the following output files:"
    file-line-max# stringbuf-open >r  r@ buffer-data-addr >r
    BEGIN  ( name-handle created-files-list-ID  r: buffer-handle buffer-addr )
	r@ file-line-max# third READ-LINE
	IF
	    bell drop
	    stringbuf-close
	    rdrop
	    r> stringbuf-close EXIT
	THEN
    WHILE
	cr r@ swap type
    REPEAT
    drop

    cr
    cr ." File  "
    created-files-list dup string@ type stringbuf-close
    ."   is a list of your output files."
    cr ." Please remove them when you're done." cr

    CLOSE-FILE drop
    stringbuf-close
    rdrop  r> stringbuf-close ;

: goodbye-file-actions ( -- )
    created-files @ IF page ELSE cr clear-line-to-end THEN
    cr ." terminating " program-name type
    .tab program-version type
    ."     Identity: " (identity) @ .
    clear-line-to-end
    ?remove-tmp-files
    ?list-created-files
    cr  clear-line-to-end ;
