\ float-input.fs

\ Words to allow user input of a float value in everydays syntax,
\ without blocking the use of the Forth text interpreter for calculations.

: dec-num? ( char -- flag )   [char] 0 [ char 9 1+ ] literal WITHIN ;

\ Count occurances of char in a string:
: count-char ( addr count char -- u )
    0 2swap
    bounds ?DO   ( char count )
	over i c@ = IF 1+ THEN
    LOOP
    nip ;

\ Check (partially) if the string contains a string which the user could
\ have entered meaning a float:
: (maybe-float?) ( addr count -- flag )
    dup 0= IF  2drop FALSE EXIT  THEN		\ empty string
 
    over c@ [char] - = IF  1 /string  THEN	\ skip '-' sign
    dup CASE \ on length			\ very short strings
	0 OF 2drop FALSE EXIT ENDOF
	1 OF
	    over c@ dec-num? IF
		2drop TRUE EXIT
	    ELSE
		2drop FALSE EXIT
	    THEN
	ENDOF
    ENDCASE

    over c@ [char] . = -rot	( starts-with-dot-flag addr count )
    0 DO			( starts-with-dot-flag addr )
	dup i + c@
	dup dec-num? IF drop ELSE
	    dup [char] E = IF  drop [char] e  THEN
	    CASE
		[char] e OF		\ 'e' and 'E'
		    over 0= IF  2drop unloop FALSE EXIT  THEN	\ looks ok
		    \ starts with dot: check char to left of 'e'
		    dup i 1- + c@ dup dec-num? IF drop ELSE
			[char] . <> IF  2drop unloop FALSE EXIT  THEN
		    THEN
		ENDOF
		[char] . OF		\ decimal point
		    i IF
			over IF  2drop unloop FALSE EXIT  THEN	\ second dot
			\ decimal point: check char to left of it
			dup i 1- + c@ dup dec-num? IF drop ELSE
			    [char] - <> IF  2drop unloop FALSE EXIT  THEN
			THEN
		    THEN
		ENDOF
		[char] - OF		\ minus sign
		    i IF				\ must be first or
			dup i 1- + c@ [char] e <> IF	\ just after 'e'
			    2drop unloop FALSE EXIT
			THEN
		    THEN
		ENDOF
		\ all other chars:
		drop 2drop unloop FALSE EXIT
	    ENDCASE
	THEN
    LOOP
    2drop TRUE ;

\ Check if the string contains a string which the user could have entered
\ in everydays syntax, meaning a float.
\ Correct floats do return FALSE, as do strings that cannot be floats anyway.
\ The heuristics are not perfect, but should be good enough, I think.
: maybe-float? ( addr count -- flag )
    2>r
    2r@ (maybe-float?) 0= IF  2rdrop FALSE EXIT  THEN

    2r@ 1- + c@  dup dec-num? IF drop ELSE	\ check last char
	CASE
	    [char] e OF ENDOF
	    [char] E OF ENDOF
	    drop 2rdrop FALSE EXIT
	ENDCASE
    THEN
    2r@ [char] . count-char 1 > IF  2rdrop FALSE EXIT  THEN	\ count '.'
    2r@ [char] - count-char 2 > IF  2rdrop FALSE EXIT  THEN	\ count '-'
    2r@ [char] e count-char  2r> [char] E count-char  +  1 > IF FALSE EXIT THEN
    TRUE ;

[UNDEFINED] (float-string) [IF]   32 STRINGBUF-HANDLE: (float-string)	[THEN]

\ Fixes a string to a correct float string in Forth syntax in place.
\ Attention: Result string can be up to 3 bytes longer than the original!
: fix-float-string ( addr count -- count' )
    over >r

    (float-string) stringbuf-empty
    over c@ [char] - = IF
	[char] - (float-string) char-cat
	1 /string
    THEN

    over c@ [char] . = IF
	[char] 0 (float-string) char-cat
    THEN

    2dup (float-string) cat

    2>r
    2r@ [char] e count-char
    2r> [char] E count-char + 0= IF
	s" e0" (float-string) cat
    THEN

    (float-string) string@  r> swap  >r r@ move
    r> ;

\ Return last word in a string of space separated words:
: last-word ( addr count -- addr' count' )
    bl -trailing-char
    dup 0= IF EXIT THEN

    BEGIN
	s"  " search WHILE
	>r 1+ r> 1-
    REPEAT ;

\ If the last word in the string looks like an attempt to input a float
\ in everydays syntax, replace it by a string in correct Forth syntax.
\ Attention: This may write up to three bytes over the end of the original!
: maybe-replace-last-as-float ( addr count -- addr count' )
    -trailing
    dup 0= IF  EXIT  THEN

    2dup last-word 2dup maybe-float? IF
	dup >r
	fix-float-string r> - +
    ELSE  2drop  THEN ;
