\ float-string.fs

[UNDEFINED] (float-string) [IF]   32 STRINGBUF-HANDLE: (float-string)	[THEN]

\ Put float representation into '(scratch-buf)' and exponent into '(scratch)'
\ and give results:
: float-scratch-represent ( r u -- exponent negative-flag success-flag )
    (scratch-buf)
    dup stringbuf-empty
    2dup string-size!
    buffer-data-addr swap represent
    third (scratch) ! ;		\ exponent for subsequent use

\ Cut trailing zeroes from the string at (scratch-buf), leaving enough
\ to display an integer with exponent 'exp'.
\ If this would result in an empty string, store '0' instead.
\ Returns the string:
: float-scratch/zeroes ( exp -- addr count' ) \ string in (scratch-buf)
    >r
    (scratch-buf) string@ [char] 0 -trailing-char
    dup 0= IF \ zero?
	drop 1
	rdrop
    ELSE
	r> max
    THEN
    dup (scratch-buf) string-size! ;

\ If the float value ASCII representation just fits into u characters return
\ it as string and TRUE, else return FALSE.
\ (use 'float-fits?' for floats fitting in a shorter string).
: float-just-fits? ( r u -- addr count TRUE | FALSE )
    dup >r
    float-scratch-represent nip
    0= IF drop rdrop FALSE EXIT THEN
    r@ = IF
	(scratch-buf) buffer-data-addr r> (float-string) cat
	(float-string) string@
	TRUE EXIT
    THEN
    rdrop FALSE ;

\ Build a ASCII representation string of a float value that can be shown as
\ mantissa only within u characters.
: build-as-mantissa ( r u -- addr count )
    >r

    fdup r@ float-scratch-represent
    0= ABORT" build-as-mantissa: Can't handle this"
    (float-string) stringbuf-empty
    IF  [char] - (float-string) char-cat  THEN
    dup float-scratch/zeroes 2drop

    \ negative exponents: add point and leading zeroes:
    dup 0< IF
	[char] . (float-string) char-cat
	0 swap DO  [char] 0 (float-string) char-cat  LOOP
	fdup r@ float-scratch-represent nip
	0= ABORT" build-as-mantissa: Can't handle this"
	dup float-scratch/zeroes 2drop
    THEN

    r> (scratch-buf) buffered-length min 0 DO
	dup i = IF  [char] . (float-string) char-cat  THEN
	(scratch-buf) buffer-data-addr i + c@ (float-string) char-cat
    LOOP drop
    fdrop
    (float-string) string@ ;

\ Return ASCII representation of the float value and TRUE if it fits as
\ mantissa only (without exponent) in *less* than u characters.
\ (use 'float-just-fits?' for floats *just* fitting).
: float-fits? ( r u -- addr count TRUE | FALSE )
    >r
    fdup r@ float-scratch-represent nip
    0= IF 2drop FALSE EXIT THEN		( r exp  r: u )

    dup float-scratch/zeroes nip	( r exp representation-length  r: u )
    swap 0 min negate +			\ space for .0[...]
    r@ <= IF
	r> build-as-mantissa
	TRUE EXIT
    THEN
    rdrop fdrop FALSE ;

\ Build a string of maximal u characters from real value r.
\ 'r' must be a real number, no infinities, no nan.
: real>string ( r u -- addr count )
    >r					( f: r  r: u=length )
    (float-string) stringbuf-empty

    fdup f0< IF				\ negative?
	s" -" (float-string) string!	\ '-' sign
	r> 1- >r			\ adjust length
    THEN

    fdup r@ float-just-fits? IF		\ just fitting integer
	fdrop rdrop EXIT
    THEN

    r> 1- >r				\ decimal point: adjust length

    fdup r@ float-fits? IF		\ fits without exponent?
	fdrop rdrop EXIT
    THEN

    \ subtract exponent string length:
    r> 1-	( r length )	\ 'e'
    (scratch) @ 1-		\ exponent (as integer)
    dup 0< IF			\ exponent sign?
	>r 1- r> negate
    THEN
    num>string nip -		\ exponent number length

    float-scratch-represent nip		( exponent success-flag )
    0= ABORT" real>string: Can't handle this number."

    (scratch-buf) buffer-data-addr c@ (float-string) char-cat	\ first number
    1-					\ new exponent
    (scratch-buf) string@ [char] 0 -trailing-char
    >r 1+ r> 1-				\ start with second character
    dup IF				\ decimal point if mantissa fraction
	[char] . (float-string) char-cat
    THEN
    (float-string) cat

    [char] e (float-string) char-cat
    dup 0< IF  [char] - (float-string) char-cat  THEN	\ exponent sign
    abs num>string (float-string) cat
    (float-string) string@ ;

\ Top level word to build a ASCII representation of maximal length 'u' from
\ a floating point value.
: float>string ( r u -- addr count )
    >r
    fdup real? IF \ real value and not infinity
	r> real>string
    ELSE \ infinity or NAN
	fdup infinity? CASE
	    1  OF
		fdrop  r> 8 > IF s" +infinity" ELSE s" +inf" THEN  EXIT
	    ENDOF
	    -1 OF
		fdrop  r> 8 > IF s" -infinity" ELSE s" +inf" THEN  EXIT
	    ENDOF
	    0  OF
		is-NaN? IF
		    rdrop s" nan" EXIT
		ELSE
		    true ABORT" float>string: Format not recognised."
		THEN
	    ENDOF
	ENDCASE
    THEN ;
